@@ -27,7 +27,7 @@ class InvalidDfError(ValueError):
     """ """
 
 
-class ExtraConditionFailedError(InvalidDfError):
+class VerificationFailedError(InvalidDfError):
     """ """
 
 
@@ -35,7 +35,7 @@ class MissingColumnError(InvalidDfError):
     """ """
 
 
-class AsymmetricDfError(InvalidDfError):
+class RowColumnMismatchError(InvalidDfError):
     """ """
 
 
@@ -53,3 +53,7 @@ class ValueNotUniqueError(ValueError):
 
 class NoValueError(ValueError):
     """ """
+
+
+class LengthMismatchError(ValueError):
+    """ """
