@@ -3,20 +3,18 @@ Defines DataFrames with convenience methods and that enforce invariants.
 """
 from __future__ import annotations
 
-from typing import Callable, Optional, Sequence, Type, Any, Mapping
+from typing import Any, Mapping, Sequence, Type, Union
 
 import pandas as pd
 
+from typeddfs._pretty_dfs import PrettyDf
 from typeddfs.base_dfs import BaseDf
 from typeddfs.df_errors import (
-    AsymmetricDfError,
-    ExtraConditionFailedError,
-    InvalidDfError,
     MissingColumnError,
     UnexpectedColumnError,
     UnexpectedIndexNameError,
+    VerificationFailedError,
 )
-from typeddfs._pretty_dfs import PrettyDf
 from typeddfs.untyped_dfs import UntypedDf
 
 
@@ -89,7 +87,7 @@ class TypedDf(BaseDf):
         df = df.reset_index()
         # remove trash columns
         df.__class__ = cls
-        df = df.drop_cols(["index", "level_0"])  # these MUST be dropped
+        df = df.drop_cols(["index", "level_0", "Unnamed: 0"])  # these MUST be dropped
         df = df.drop_cols(cls.columns_to_drop())
         # now let's convert the dtypes
         for c, dt in cls.auto_dtypes().items():
@@ -113,6 +111,18 @@ class TypedDf(BaseDf):
         for c in list(cls.required_columns()) + list(cls.reserved_columns()):
             if c not in new_columns and c in df.columns:
                 new_columns.append(c)
+        # set the index/column series name(s)
+        df: BaseDf = df
+        col_series = cls.column_series_name()
+        if col_series is not False:
+            if col_series is True:
+                col_series = None
+            df.columns.name = col_series
+        ind_series = cls.index_series_name()
+        if df.is_multindex() and ind_series is not False:
+            if ind_series is True:
+                ind_series = None
+            df.index.name = ind_series
         # this lets us keep whatever extra columns
         df = df.cfirst(new_columns)
         # call post-processing
@@ -124,6 +134,26 @@ class TypedDf(BaseDf):
         df.__class__ = cls
         return df
 
+    @classmethod
+    def new_df(cls, *, reserved: Union[bool, Sequence[str]] = False) -> __qualname__:
+        """
+        Returns a DataFrame that is empty but has the correct columns and indices.
+
+        Arguments:
+            reserved: Include reserved index/column names as well as required.
+                      If True, adds all of ``cls.reserved_index_names()`` and ``cls.reserved_columns``.
+                      You can also specify the exact list of columns and index names.
+
+        Raises:
+            InvalidDfError: If a function in ``verifications`` fails (returns False or a string).
+        """
+        if reserved:
+            req = cls.known_names()
+        else:
+            req = [*cls.required_index_names(), *cls.required_columns()]
+        df = pd.DataFrame({r: [] for r in req})
+        return cls.convert(df)
+
     def untyped(self) -> UntypedDf:
         """
         Makes a copy that's an UntypedDf.
@@ -157,16 +187,23 @@ class TypedDf(BaseDf):
             return self.__class__.convert(df)
 
     @classmethod
-    def is_valid(cls, df: pd.DataFrame) -> bool:
+    def column_series_name(cls) -> Union[bool, None, str]:
         """
-        Returns True if all of the required conditions pass.
-        This is provided as a sanity check only: It should always return True.
+        Returns a value that will be forced into ``df.columns.name`` on calling ``convert``.
+        If ``None``, will set ``df.columns.name = None``.
+        If ``False``, will not set. (``True`` is treated the same as ``None``.)
         """
-        try:
-            cls.convert(df)
-            return True
-        except InvalidDfError:
-            return False
+        return None
+
+    @classmethod
+    def index_series_name(cls) -> Union[bool, None, str]:
+        """
+        Returns a value that will be forced into ``df.index.name`` on calling ``convert``,
+        *only if* the DataFrame is multi-index.
+        If ``None``, will set ``df.index.name = None`` if ``df.index.names != [None]``.
+        If ``False``, will not set. (``True`` is treated the same as ``None``.)
+        """
+        return None
 
     @classmethod
     def more_columns_allowed(cls) -> bool:
@@ -178,7 +215,7 @@ class TypedDf(BaseDf):
     @classmethod
     def more_indices_allowed(cls) -> bool:
         """
-        Returns whether the DataFrame allows index names that are not reserved or required.
+        Returns whether the DataFrame allows index levels that are neither reserved nor required.
         """
         return True
 
@@ -206,7 +243,7 @@ class TypedDf(BaseDf):
     @classmethod
     def reserved_index_names(cls) -> Sequence[str]:
         """
-        Returns the list of reserved (optional) index names.
+        Returns the list of reserved (optional) index levels.
         """
         return []
 
@@ -221,7 +258,7 @@ class TypedDf(BaseDf):
     @classmethod
     def known_index_names(cls) -> Sequence[str]:
         """
-        Returns all index names that are required or reserved.
+        Returns all index levels that are required or reserved.
         The sort order positions required columns first.
         """
         return [*cls.required_index_names(), *cls.reserved_index_names()]
@@ -251,13 +288,6 @@ class TypedDf(BaseDf):
         """
         return {}
 
-    @classmethod
-    def must_be_symmetric(cls) -> bool:
-        """
-        Returns whether the (single only) index values must match the column names.
-        """
-        return False
-
     @classmethod
     def columns_to_drop(cls) -> Sequence[str]:
         """
@@ -266,53 +296,26 @@ class TypedDf(BaseDf):
         """
         return []
 
-    @classmethod
-    def post_processing(cls) -> Optional[Callable[[BaseDf], Optional[BaseDf]]]:
-        """
-        A function to be called at the final stage of ``convert``.
-        It is called immediately before ``extra_conditions`` are checked.
-        The function takes a copy of the input ``BaseDf`` and returns a new copy.
-
-        Note:
-            Although a copy is passed as input, the function should not modify it.
-            Technically, doing so will cause problems only if the DataFrame's internal values
-            are modified. The value passed is a *shallow* copy (see ``pd.DataFrame.copy``).
-        """
-        return None
-
-    @classmethod
-    def extra_conditions(cls) -> Sequence[Callable[[BaseDf], Optional[str]]]:
-        """
-        Additional requirements for the DataFrame to be conformant.
-
-        Returns:
-            A sequence of conditions that map the DF to None if the condition passes,
-            or the string of an error message if it fails
-        """
-        return []
-
     @classmethod
     def _check(cls, df) -> None:
         cls._check_has_required(df)
         cls._check_has_unexpected(df)
-        if cls.must_be_symmetric():
-            cls._check_symmetric(df)
-        for req in cls.extra_conditions():
+        for req in cls.verifications():
             value = req(df)
             if value is not None:
-                raise ExtraConditionFailedError(value)
+                raise VerificationFailedError(value)
 
     @classmethod
     def _check_has_required(cls, df: pd.DataFrame) -> None:
         for c in set(cls.required_index_names()):
             if c not in set(df.index.names):
                 raise MissingColumnError(
-                    f"Missing index name {c} (indices are: {set(df.index.names)}; columns are: {set(df.columns.names)}))"
+                    f"Missing index name {c} (indices are: {set(df.index.names)}; cols are: {set(df.columns.names)}))"
                 )
         for c in set(cls.required_columns()):
             if c not in set(df.columns):
                 raise MissingColumnError(
-                    f"Missing column {c} (columns are: {set(df.columns.names)}; indices are: {set(df.index.names)})"
+                    f"Missing column {c} (cols are: {set(df.columns.names)}; indices are: {set(df.index.names)})"
                 )
 
     @classmethod
@@ -327,17 +330,6 @@ class TypedDf(BaseDf):
                 if c not in cls.required_index_names() and c not in cls.reserved_index_names():
                     raise UnexpectedIndexNameError(f"Unexpected index name {c}")
 
-    @classmethod
-    def _check_symmetric(cls, df: pd.DataFrame) -> None:
-        if isinstance(df.index, pd.MultiIndex):
-            raise AsymmetricDfError(
-                f"The {cls.__name__} cannot be symmetric because it's multi-index"
-            )
-        if list(df.index) != list(df.columns):
-            raise AsymmetricDfError(
-                f"The indices are {list(df.index)} but the rows are {list(df.columns)}"
-            )
-
     @classmethod
     def _lines_files_apply(cls) -> bool:
         return len(cls.known_names()) == 1
